// This may look like C code, but it is really -*- C++ -*-
//
//	Verify construction of fractal maps and show them off
//
// $Id: vfractal_map.cc,v 2.2 1996/07/15 19:07:30 oleg Exp oleg $

#include "./image.h"
#include "fractal_map.h"
#include "filter.h"
#include "iostream.h"
#include "MyImageViewer.h"

static void test_nonoise(const card order)
{
  cout << "\nTest constructing a map of order " << order << 
	  " without any noise\n" << endl;

  class ZeroNoise : public FractalMap
  {
  public:
    ZeroNoise(const card order) : FractalMap(order) {}
    ZeroNoise(const card order,const card smoothness,
  	      const card starting_scale_log)
      : FractalMap(order,smoothness,starting_scale_log) {}
    inline int get_noise(const card scale) const { return 0; }
  };

  IMAGE map = ZeroNoise(order);
  const GRAY pixel_mid_val = 1<<(map.q_depth()-1);
  verify_pixel_value(map,pixel_mid_val);
  IMAGE map1 = ZeroNoise(order,0,2);
  verify_pixel_value(map1,pixel_mid_val);
  cout << "\nDone" << endl;
}

static void test_norandom(void)
{
  const card order = 2;
  cout << "\nTest constructing a map of order " << order << 
	  " with non-random noise\n" << endl;

  class NoRNoise : public FractalMap
  {
  public:
    NoRNoise(const card order, const card smoothness,
  	     const card starting_scale_log,
	     const card bits_per_pixel)
      : FractalMap(order,smoothness,starting_scale_log,bits_per_pixel)
      { cout << "\tsmoothness " << smoothness << ", starting scale log"
             << starting_scale_log << ",depth " << bits_per_pixel << endl; }
    inline int get_noise(const card scale) const { return scale/2; }
  };

  {
    IMAGE map = NoRNoise(order,0,1,7);
    verify_pixel_value(map,96);
  }
  {
    IMAGE map = NoRNoise(order,0,2,7);
    verify_pixel_value(map,112);
  }
  {
    IMAGE map = NoRNoise(order,1,2,7);
    verify_pixel_value(map,96);
  }
  
#if 0
  map.print("map");

  IMAGE expected(map);
  expected = 2; expected(0,0) = 0;
  expected.square_of(3,rowcol(1,1)) = 4;
  assert( map == expected );
#endif

  cout << "\nDone" << endl;
}


class FractalMapWindow : public BWindow
{
  BLooper& parent;
  IMAGE map;
  
  class PIntItem
  {
    const char * const name;
    short value;
   public:
    PIntItem(const char _name [], const int default_value)
    	: name(_name), value(default_value) {}
    operator int (void) const		{ return value; }
    const PIntItem& operator << (const BMessage& db);
  };
  
  struct Porder : public PIntItem
  {
    Porder(const int value = 8)
    	: PIntItem("order",value) {}
  };
  struct Psmoothness : public PIntItem
  {
    Psmoothness(const int value = 0)
    	: PIntItem("smoothness",value) {}
  };
  struct Pstarting_scale_log : public PIntItem
  {
    Pstarting_scale_log(const int value = 0)
    	: PIntItem("ss_log",value) {}
  };

public:

   enum { FWPMessage = 100 };
   
   FractalMapWindow(const BMessage& params, BLooper& _parent);
   bool QuitRequested(void)
     { parent.PostMessage(B_QUIT_REQUESTED); return true; }
};

			// Parameters database access
const FractalMapWindow::PIntItem&
FractalMapWindow::PIntItem::operator << (const BMessage& parms)
{
  if( parms.HasInt16(name) )
  {
    assert( parms.FindInt16(name,&value) == B_OK );
  }
  return *this;
}

FractalMapWindow::FractalMapWindow(const BMessage& params, BLooper& _parent)
   : BWindow(BRect(50,50,80,80),		// just to construct the window; will
   									// be resized later, after map is
   									// constructed
   			  "A Fractal Map",B_TITLED_WINDOW,
   			  B_NOT_ZOOMABLE|B_NOT_RESIZABLE|B_NOT_MINIMIZABLE),
   	parent(_parent),
   	map(FractalMap(Porder()<<params,
   				   Psmoothness()<<params,
   				   Pstarting_scale_log()<<params))
   	{
   	  ResizeTo(map.q_ncols(),map.q_nrows());	// Set the proper dimensions
   	  Show();
   	  AddChild(new MyImageView(map));
   	  //SetTarget(&parent);
   	}

#if 0
				// Show off maps
static void demo_maps(void)
{
  class GaussNoise : public FractalMap
  {
  public:
    GaussNoise(const card order, const card smoothness,
  	     const card starting_scale_log)
      : FractalMap(order,smoothness,starting_scale_log) {}
				// Well-known Gaussian random number generator
				// Note rand() returns a number [0,2^15-1],
				// that is, within [0,1] with a 15-bit implicit
				// fraction
    inline int get_noise(const card scale) const {
      long sum = 0;
      for(register int i=0; i<12; i++)
	sum += rand();			// keep the result within 
      return (scale * (sum-(6<<15)))>>17; } // [-scale/2,scale/2]
  };

  WindowDestroyer window_destroyer;
  
  for(;;)
  {
    cout << "Enter a map's order (log2 of map's dimension) (0 to quit)" 
      << endl;
    int order;
    cin >> order;
    if( order <= 1 )
     return;
    cout << "Select random number generator: 0 - Uniform, 1 - Gaussian"
         << endl;
    int type;
    cin >> type;
    cout << "Enter smoothness and the starting scale log (both can be zero)"
         << endl;
    int smoothness, starting_scale_log;
    cin >> smoothness >> starting_scale_log;
    IMAGE map = type == 0 ? 
      (LazyImage&)FractalMap(order,smoothness,starting_scale_log) :
      (LazyImage&)FractalMap(order,smoothness,starting_scale_log);
    cout << "Select post-processing option" << endl;
    cout << "\t0 - no postprocessing" << endl;
    cout << "\t1 - apply blurring" << endl;
    cout << "\t2 - smooth non-deterministically" << endl;
    cin >> type;
    if( type == 1 )
      filter(map).conv(conv_kernel(1,1,1,CommonDenom(3))).
	clip_to_intensity_range();
    else if( type == 2 )
      filter(map).nondet_average(50);

    char buffer[70];
    sprintf(buffer,"smoothness %d, starting_scale_log %d",
    	    smoothness,starting_scale_log);
    window_destroyer.add(new MyImageWindow(map,buffer));
//    MyImageViewer image_viewer(map,buffer);
//    map.write_pgm("/tmp/abc",buffer);
//    system("/usr/local/bin/X11/xv /tmp/abc");
//    map.display(buffer);
  }
}

#endif

int main(void)
{
  test_nonoise(4); 
  test_nonoise(9); 
  test_norandom();

  srand(time(0));
  rand(); rand(); rand(); rand();

class MyImageViewer : public BApplication
{
  FractalMapWindow * window;
public:
	    MyImageViewer(void)
	    : BApplication("application/x-vnd.x-vfractal-map")
	    { }
	    void ReadyToRun(void)
	    {
	      window = new FractalMapWindow(
	      		BMessage(FractalMapWindow::FWPMessage),*this);
	    }
	    
};
MyImageViewer().Run();

printf("\nDone\n\n");
return 0;
 
}
